from newnewid.clock.uuid_clock import UUIDClock
from newnewid.counter.counter import Counter
from newnewid.draft_peabody_dispatch_new_uuid_format_01.uuid6_generator import (
    UUID6Generator,
    uuid6,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_01.uuid7_generator import (
    UUID7Generator,
    uuid7,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_01.uuid8_generator import (
    UUID8Generator,
    uuid8,
)
from newnewid.exception.backward_uuid_exception import BackwardUUIDException
from newnewid.random.pseudo_random_binary_generator import PseudoRandomBinaryGenerator
from newnewid.util.specloader import UUIDSpec
from newnewid.uuidgenerator.gregorian_based_uuid_generator import MacAddressGenerator
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator

__all__ = [
    "UUIDGenerator",
    "uuid6",
    "UUID6Generator",
    "uuid7",
    "UUID7Generator",
    "uuid8",
    "UUID8Generator",
    "BackwardUUIDException",
    "PseudoRandomBinaryGenerator",
    "Counter",
    "UUIDClock",
    "UUIDSpec",
    "MacAddressGenerator",
]



from datetime import datetime, timedelta
from typing import Any, Dict, Literal, Optional, Tuple
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.counter.counter import Counter
from newnewid.random.pseudo_random_binary_generator import PseudoRandomBinaryGenerator
from newnewid.random.pseudo_random_generator import PseudoRandomGenerator
from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.clock_based_uuid_generator import ClockBasedUUIDGenerator

Precision = Literal["milli", "micro", "nano"]


class UUID7Generator(ClockBasedUUIDGenerator):
    """UUIDv7 generator class.

    UUIDv7 is a 128-bit UUID that is based on Epoch time and is sortable.

    There are three variations of UUIDv7 depending on the granularity of the epoch time.

    * Millisecond variation

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                            unixts                             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |unixts |         msec          |  ver  |          seq          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |var|                         rand                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                             rand                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    * Microsecond variation

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                            unixts                             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |unixts |         usec          |  ver  |         usec          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |var|             seq           |            rand               |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                             rand                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    * Nanosecond variation

       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                            unixts                             |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |unixts |         nsec          |  ver  |         nsec          |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |var|             nsec          |      seq      |     rand      |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                             rand                              |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    def __init__(
        self,
        precision: Precision,
        raise_exception_on_backward: bool = False,
        clock: Optional[UUIDClock] = None,
        pseudo_random_generator: Optional[PseudoRandomGenerator] = None,
        last_uuid: Optional[UUID] = None,
    ) -> None:
        """Create UUID7 generator.

        Args:
            precision (Precision): Precision of the UUID7.
            raise_exception_on_backward (bool, optional): Raise exception on backward clock if True. Otherwise, wait for the clock to move forward. Defaults to False.
            clock (Optional[UUIDClock], optional): Clock. Defaults to None.
            pseudo_random_generator (Optional[PseudoRandomGenerator], optional): Pseudo random generator. Defaults to None.
            last_uuid (Optional[UUID], optional): Last UUID. Defaults to None.
        """
        super().__init__(raise_exception_on_backward, clock, pseudo_random_generator, last_uuid)

        self.precision: Precision = precision

        if precision == "milli":
            self.cbl = 12
            self.random_bits = 52
        elif precision == "micro":
            self.cbl = 14
            self.random_bits = 48
        else:
            self.cbl = 8
            self.random_bits = 40

        last_timestamp: Optional[int] = None
        last_counter: Optional[int] = None
        if last_uuid:
            assert last_uuid.version == 7, f"last_uuid must be version 7, not {last_uuid.version}"
            (last_timestamp, last_counter) = self.__class__._parse_last_uuid(last_uuid, precision)

        self._counter = Counter(
            bits_length=self.cbl,
            max_increment_bits_length=1,
            pseudo_random_generator=self._pseudo_random_generator,
            initial_timestamp=last_timestamp,
            initial_counter=last_counter,
        )
        self._pseudo_random_binary_generator = PseudoRandomBinaryGenerator(
            self.random_bits, self._pseudo_random_generator
        )

    @nodoc
    def timestamp(self, clock: UUIDClock) -> int:
        if self.precision == "milli":
            return clock.epoch_36_bits_seconds_with_12_bits_milli_seconds()
        elif self.precision == "micro":
            return clock.epoch_36_bits_seconds_with_24_bits_micro_seconds()
        elif self.precision == "nano":
            return clock.epoch_36_bits_seconds_with_38_bits_nano_seconds()
        else:
            raise ValueError(f"Invalid precision: {self.precision}")

    @property
    @nodoc
    def clock_bits_length(self) -> int:
        if self.precision == "milli":
            return 48
        elif self.precision == "micro":
            return 60
        elif self.precision == "nano":
            return 74
        else:
            raise ValueError(f"Invalid precision: {self.precision}")

    @property
    @nodoc
    def least_seconds(self) -> float:
        if self.precision == "milli":
            return 0.001
        elif self.precision == "micro":
            return 0.000_001
        elif self.precision == "nano":
            return 0.000_000_001
        else:
            raise ValueError(f"Invalid precision: {self.precision}")

    @nodoc
    def generate_impl(
        self,
        timestamp: int,
    ) -> UUID:
        precision = self.precision

        var = 0b10
        ver = 0b0111

        seq = self._counter.get_next(timestamp)

        rand = self._pseudo_random_binary_generator.generate()

        if precision == "milli":
            # 36 bits
            unixts = (timestamp >> 12) & 0x000F_FFFF_FFFF

            # 12 bits
            msec = timestamp & 0x0FFF

            uuid_int = (
                (unixts << 92) | (msec << 80) | (ver << 76) | (seq << 64) | (var << 62) | rand
            )
        elif precision == "micro":
            # 36 bits
            unixts = (timestamp >> 24) & 0x000F_FFFF_FFFF

            # 12 bits
            usec_a = (timestamp >> 12) & 0x0FFF

            # 12 bits
            usec_b = timestamp & 0x0FFF

            uuid_int = (
                (unixts << 92)
                | (usec_a << 80)
                | (ver << 76)
                | (usec_b << 64)
                | (var << 62)
                | (seq << 48)
                | rand
            )
        else:
            # 36 bits
            unixts = (timestamp >> 38) & 0x000F_FFFF_FFFF

            # 12 bits
            nsec_a = (timestamp >> 26) & 0x0FFF

            # 12 bits
            nsec_b = (timestamp >> 14) & 0x0FFF

            # 14 bits
            nsec_c = timestamp & 0x3FFF

            uuid_int = (
                (unixts << 92)
                | (nsec_a << 80)
                | (ver << 76)
                | (nsec_b << 64)
                | (var << 62)
                | (nsec_c << 48)
                | (seq << 40)
                | rand
            )

        return UUID(int=uuid_int)

    @classmethod
    def parse(cls, uuid: UUID, precision: Precision, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUID.

        Args:
            uuid (UUID): UUID.
            precision (Precision): Precision.

        Returns:
            Dict[str, Any]: Parsed UUID.
        """
        unixts = (uuid.int >> 92) & 0x000F_FFFF_FFFF
        if precision == "milli":
            msec = (uuid.int >> 80) & 0x0FFF
            version = (uuid.int >> 76) & 0x000F
            seq = (uuid.int >> 64) & 0x0FFF
            variant = (uuid.int >> 62) & 0x0003
            rand = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

            time = datetime.fromtimestamp(unixts) + timedelta(milliseconds=msec)
            return {
                "unixts": unixts,
                "msec": msec,
                "ver": str(version),
                "seq": seq,
                "variant": variant,
                "rand": rand,
                "time": time.isoformat(),
            }
        elif precision == "micro":
            usec_a = (uuid.int >> 80) & 0x0FFF
            version = (uuid.int >> 76) & 0x000F
            usec_b = (uuid.int >> 64) & 0x0FFF
            variant = (uuid.int >> 62) & 0x0003
            seq = (uuid.int >> 48) & 0x3FFF
            rand = uuid.int & 0xFFFF_FFFF_FFFF

            usec = (usec_a << 12) | usec_b
            time = datetime.fromtimestamp(unixts) + timedelta(microseconds=usec)
            return {
                "unixts": unixts,
                "usec_a": usec_a,
                "ver": str(version),
                "usec_b": usec_b,
                "variant": variant,
                "rand": rand,
                "time": time.isoformat(),
            }
        else:
            nsec_a = (uuid.int >> 80) & 0x0FFF
            version = (uuid.int >> 76) & 0x000F
            nsec_b = (uuid.int >> 64) & 0x0FFF
            variant = (uuid.int >> 62) & 0x0003
            nsec_c = (uuid.int >> 48) & 0x3FFF
            seq = (uuid.int >> 40) & 0x00FF
            rand = uuid.int & 0x00FF_FFFF_FFFF

            nsec = (nsec_a << 26) | (nsec_b << 14) | nsec_c
            time = datetime.fromtimestamp(unixts) + timedelta(microseconds=nsec // 1000)
            return {
                "unixts": unixts,
                "nsec_a": nsec_a,
                "ver": str(version),
                "nsec_b": nsec_b,
                "variant": variant,
                "nsec_c": nsec_c,
                "seq": seq,
                "rand": rand,
                "time": time.isoformat(),
                "nsec": nsec,
            }

    @classmethod
    def _parse_last_uuid(cls, last_uuid: UUID, precision: Precision) -> Tuple[int, int]:
        parsed = cls.parse(last_uuid, precision)

        if precision == "milli":
            unixts = parsed["unixts"]
            msec = parsed["msec"]

            time = unixts << 12 | msec
        elif precision == "micro":
            unixts = parsed["unixts"]
            usec_a = parsed["usec_a"]
            usec_b = parsed["usec_b"]

            time = unixts << 24 | usec_a << 12 | usec_b
        elif precision == "nano":
            unixts = parsed["unixts"]
            nsec_a = parsed["nsec_a"]
            nsec_b = parsed["nsec_b"]
            nsec_c = parsed["nsec_c"]

            time = unixts << 38 | nsec_a << 26 | nsec_b << 14 | nsec_c
        else:
            raise ValueError("Logic error")

        return time, parsed["seq"]


_uuid7_generator_milli = UUID7Generator("milli")
_uuid7_generator_micro = UUID7Generator("micro")
_uuid7_generator_nano = UUID7Generator("nano")


def uuid7(precision: Precision = "micro") -> UUID:
    """Generate UUIDv7.

    Args:
        precision (Precision, optional): Precision. Defaults to "micro".

    Returns:
        UUID: UUIDv7.
    """
    if precision == "milli":
        return _uuid7_generator_milli.generate()
    elif precision == "micro":
        return _uuid7_generator_micro.generate()
    else:
        return _uuid7_generator_nano.generate()



import uuid
from abc import ABCMeta, abstractmethod
from typing import Optional
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.counter.counter import Counter
from newnewid.random.pseudo_random_binary_generator import PseudoRandomBinaryGenerator
from newnewid.random.pseudo_random_generator import PseudoRandomGenerator
from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.clock_based_uuid_generator import ClockBasedUUIDGenerator


class MacAddressGenerator(PseudoRandomBinaryGenerator):
    """MAC address generator."""

    def __init__(self) -> None:
        super().__init__(48, PseudoRandomGenerator())

    @nodoc
    def generate(self) -> int:
        return uuid.getnode()


class GregorianBasedUUIDGenerator(ClockBasedUUIDGenerator, metaclass=ABCMeta):
    """UUID generator that uses Gregorian calendar."""

    def __init__(
        self,
        uses_mac_address: bool,
        raise_exception_on_backward: bool = False,
        clock: Optional[UUIDClock] = None,
        pseudo_random_generator: Optional[PseudoRandomGenerator] = None,
        last_uuid: Optional[UUID] = None,
    ) -> None:
        """Create GregorianBasedUUIDGenerator.

        Args:
            uses_mac_address (bool): Uses MAC address if True. Otherwise, uses pseudo random generator.
            raise_exception_on_backward (bool, optional): Raise exception when backward UUID is generated if True. Otherwise, sleep until next UUID can be generated. Defaults to False.
            clock (Optional[UUIDClock], optional): Clock. Defaults to None.
            pseudo_random_generator (Optional[PseudoRandomGenerator], optional): Pseudo random generator. Defaults to None.
            last_uuid (Optional[UUID], optional): Last generated UUID. Defaults to None.
        """
        super().__init__(raise_exception_on_backward, clock, pseudo_random_generator, last_uuid)

        last_timestamp: Optional[int] = None
        last_counter: Optional[int] = None
        if last_uuid:
            assert last_uuid.version == 6, f"last_uuid must be version 6, not {last_uuid.version}"
            parsed = self.__class__.parse(last_uuid)
            (time_high, time_mid, time_low) = (
                parsed["time_high"],
                parsed["time_mid"],
                parsed["time_low"],
            )

            last_timestamp = time_high << 20 | time_mid << 4 | time_low
            last_counter = parsed["clock_seq"]

        self._counter = Counter(
            bits_length=14,
            max_increment_bits_length=1,
            pseudo_random_generator=self._pseudo_random_generator,
            initial_timestamp=last_timestamp,
            initial_counter=last_counter,
        )
        if uses_mac_address:
            self._pseudo_random_binary_generator = MacAddressGenerator()
        else:
            self._pseudo_random_binary_generator = PseudoRandomBinaryGenerator(
                48, self._pseudo_random_generator
            )

    @nodoc
    def timestamp(self, clock: UUIDClock) -> int:
        return clock.gregorian_100_nano_seconds()

    @property
    @nodoc
    def clock_bits_length(self) -> int:
        return 60

    @nodoc
    @property
    def least_seconds(self) -> float:
        return 0.000_000_1

    @nodoc
    def generate_impl(self, timestamp: int) -> UUID:
        # timestamp is 60 bits

        # 4 bits
        ver = 6

        # 2 bits
        var = 0b10

        # 14 bits
        clock_seq = self._counter.get_next(timestamp)

        # 48 bits
        node = self._pseudo_random_binary_generator.generate()

        return self.build(timestamp, ver, var, clock_seq, node)

    @abstractmethod
    def build(
        self,
        time: int,
        ver: int,
        var: int,
        clock_seq: int,
        node: int,
    ) -> UUID:
        """Build UUID.

        Args:
            time (int): Timestamp.
            ver (int): Version
            var (int): Variant
            clock_seq (int): Clock sequence (Counter).
            node (int): Node.

        Returns:
            UUID: UUID.
        """



from uuid import UUID


class BackwardUUIDException(Exception):
    """Exception raised when backward UUID is generated."""

    def __init__(self, last_uuid: UUID, backward_uuid: UUID, *args) -> None:
        """Create a new BackwardUUIDException.

        Args:
            last_uuid (UUID): Last generated UUID.
            backward_uuid (UUID): Backward UUID.
        """
        super().__init__(*args)
        self.last_uuid = last_uuid
        self.backwarded_uuid = backward_uuid



import secrets


class PseudoRandomGenerator:
    """Pseudo random generator."""

    def generate(self, bits: int, use: str) -> int:
        """Generate pseudo random bits.

        CSPRNG is used for the generation.

        Args:
            bits (int): Number of bits to generate.
            use (str): Use of the generated bits. This is used for mocking in test classes.

        Returns:
            int: Pseudo random bits.
        """
        return secrets.randbits(bits)



from typing import Any, Dict
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.util.specloader import UUIDSpec, load_modules
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator


class _OtherUUIDParser:
    @classmethod
    def parse_uuid1(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUIDv1."""
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                           time_low                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |           time_mid            |  ver  |       time_high       |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |var|         clock_seq         |             node              |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                              node                             |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

        # 32 bits
        time_low = (uuid.int >> 96) & 0xFFFF_FFFF
        # 16 bits
        time_mid = (uuid.int >> 80) & 0xFFFF
        # 4 bits
        ver = (uuid.int >> 76) & 0xF
        # 12 bits
        time_high = (uuid.int >> 64) & 0xFFF
        # 2 bits
        var = (uuid.int >> 62) & 0x3
        # 14 bits
        clock_seq = (uuid.int >> 48) & 0x3FFF
        # 48 bits
        node = uuid.int & 0xFFFF_FFFF_FFFF

        gregorian_100_nano_seconds = (time_high << 48) | (time_mid << 32) | time_low
        time, epoch_nano_fraction = UUIDClock.to_datetime_from_gregorian_100_nano_seconds(
            gregorian_100_nano_seconds
        )

        return {
            "time_low": time_low,
            "time_mid": time_mid,
            "ver": str(ver),
            "time_high": time_high,
            "var": var,
            "clock_seq": clock_seq,
            "node": node,
            "gregorian_100_nano_seconds": gregorian_100_nano_seconds,
            "time": time.isoformat(),
            "epoch_nano_fraction": epoch_nano_fraction,
        }

    @classmethod
    def parse_uuid3(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUIDv3."""
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                            md5_high                           |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |          md5_high             |  ver  |       md5_mid         |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |var|                        md5_low                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                            md5_low                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # 48 bits
        md5_high = (uuid.int >> 80) & 0xFFFF_FFFF_FFFF
        # 4 bits
        ver = (uuid.int >> 76) & 0xF
        # 12 bits
        md5_mid = (uuid.int >> 64) & 0xFFF
        # 2 bits
        var = (uuid.int >> 62) & 0x3
        # 62 bits
        md5_low = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

        md5 = (md5_high << 64) | (md5_mid << 48) | md5_low

        return {
            "md5_high": md5_high,
            "ver": str(ver),
            "md5_mid": md5_mid,
            "var": var,
            "md5_low": md5_low,
            "md5": md5,
        }

    @classmethod
    def parse_uuid4(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUIDv4."""
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                           random_a                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |          random_a             |  ver  |       random_b        |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |var|                       random_c                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                           random_c                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # 48 bits
        random_a = (uuid.int >> 80) & 0xFFFF_FFFF_FFFF
        # 4 bits
        ver = (uuid.int >> 76) & 0xF
        # 12 bits
        random_b = (uuid.int >> 64) & 0xFFF
        # 2 bits
        var = (uuid.int >> 62) & 0x3
        # 62 bits
        random_c = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

        return {
            "random_a": random_a,
            "ver": str(ver),
            "random_b": random_b,
            "var": var,
            "random_c": random_c,
        }

    @classmethod
    def parse_uuid5(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUIDv5."""
        #     0                   1                   2                   3
        #     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                           sha1_high                           |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |         sha1_high             |  ver  |      sha1_mid         |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |var|                       sha1_low                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        #    |                           sha1_low                            |
        #    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        # 48 bits
        sha1_high = (uuid.int >> 80) & 0xFFFF_FFFF_FFFF
        # 4 bits
        ver = (uuid.int >> 76) & 0xF
        # 12 bits
        sha1_mid = (uuid.int >> 64) & 0xFFF
        # 2 bits
        var = (uuid.int >> 62) & 0x3
        # 62 bits
        sha1_low = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

        sha1 = (sha1_high << 64) | (sha1_mid << 48) | sha1_low

        return {
            "sha1_high": sha1_high,
            "ver": str(ver),
            "sha1_mid": sha1_mid,
            "var": var,
            "sha1_low": sha1_low,
            "sha1": sha1,
        }


def parse(uuid: UUID, spec: UUIDSpec = "latest", **kwargs) -> Dict[str, Any]:
    """Parse UUID.

    Args:
        uuid (UUID): UUID.
        spec (UUIDSpec, optional): UUID spec. Defaults to "latest".

    Returns:
        Dict[str, Any]: Parsed UUID.
    """
    uuid_int = uuid.int
    version = UUIDGenerator.get_version(uuid)

    module = load_modules(spec)
    try:
        if uuid_int == 0:
            parse_func = module.NilUUIDGenerator.parse  # type: ignore # noqa: F821
        elif uuid_int == 0xFFFF_FFFF_FFFF_FFFF_FFFF_FFFF_FFFF_FFFF:
            parse_func = module.MaxUUIDGenerator.parse  # type: ignore # noqa: F821
        elif version == "1":
            parse_func = _OtherUUIDParser.parse_uuid1
        elif version == "3":
            parse_func = _OtherUUIDParser.parse_uuid3
        elif version == "4":
            parse_func = _OtherUUIDParser.parse_uuid4
        elif version == "5":
            parse_func = _OtherUUIDParser.parse_uuid5
        elif version == "6":
            parse_func = module.UUID6Generator.parse  # type: ignore # noqa: F821
        elif version == "7":
            parse_func = module.UUID7Generator.parse  # type: ignore # noqa: F821
        elif version == "8":
            parse_func = module.UUID8Generator.parse  # type: ignore # noqa: F821
        else:
            raise ValueError(f"Unsupported version: {version}")
    except UnboundLocalError:
        raise ValueError(f"Unsupported version: {version}")

    return parse_func(uuid, **kwargs)  # type: ignore # noqa: F821



from uuid import (
    NAMESPACE_DNS,
    NAMESPACE_OID,
    NAMESPACE_URL,
    NAMESPACE_X500,
    RESERVED_FUTURE,
    RESERVED_MICROSOFT,
    RESERVED_NCS,
    RFC_4122,
    UUID,
    SafeUUID,
    getnode,
    uuid1,
    uuid3,
    uuid4,
    uuid5,
)

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.counter.counter import Counter
from newnewid.draft_ietf_uuidrev_rfc4122bis_00.nil_uuid_generator import (
    NilUUIDGenerator,
    nil_uuid,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_01.uuid6_generator import (
    UUID6Generator,
    uuid6,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_03.max_uuid_generator import (
    MaxUUIDGenerator,
    max_uuid,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_03.uuid7_generator import (
    METHOD_0_NO_COUNTER,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_26,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_42,
    METHOD_2_MONOTONIC_RANDOM_62_BITS,
    UUID7Generator,
    UUID7Option,
    ulid_compatible,
    uuid7,
)
from newnewid.draft_peabody_dispatch_new_uuid_format_03.uuid8_generator import (
    UUID8Generator,
    uuid8,
)
from newnewid.exception.backward_uuid_exception import BackwardUUIDException
from newnewid.random.pseudo_random_binary_generator import PseudoRandomBinaryGenerator
from newnewid.random.pseudo_random_generator import PseudoRandomGenerator
from newnewid.util.specloader import UUIDSpec
from newnewid.uuidgenerator.gregorian_based_uuid_generator import MacAddressGenerator
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator

__all__ = [
    # Compatibility
    "NAMESPACE_DNS",
    "NAMESPACE_OID",
    "NAMESPACE_URL",
    "NAMESPACE_X500",
    "RESERVED_FUTURE",
    "RESERVED_MICROSOFT",
    "RESERVED_NCS",
    "RFC_4122",
    "UUID",
    "SafeUUID",
    "getnode",
    "uuid1",
    "uuid3",
    "uuid4",
    "uuid5",
    # New UUID
    "UUIDGenerator",
    "max_uuid",
    "MaxUUIDGenerator",
    "nil_uuid",
    "NilUUIDGenerator",
    "uuid6",
    "UUID6Generator",
    "METHOD_0_NO_COUNTER",
    "METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12",
    "METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_26",
    "METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_42",
    "METHOD_2_MONOTONIC_RANDOM_62_BITS",
    "UUID7Generator",
    "UUID7Option",
    "ulid_compatible",
    "uuid7",
    "uuid8",
    "UUID8Generator",
    "BackwardUUIDException",
    "PseudoRandomBinaryGenerator",
    "Counter",
    "UUIDClock",
    "PseudoRandomGenerator",
    "UUIDSpec",
    "MacAddressGenerator",
]



from typing import Any, Dict
from uuid import UUID

from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator


class UUID8Generator(UUIDGenerator):
    """UUIDv8 generator class.

    UUIDv8 is a 128-bit UUID that is based on any time system and is sortable.

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                          timestamp_32                         |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |           timestamp_48        |  ver  |      time_or_seq      |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |var|  seq_or_node  |          node                             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                              node                             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    You can extend the timestamp part of the UUIDv8 by using `time_or_seq` and the
    node part (pseudo-random value) by using `seq_or_node`. You can then determine
    the sequence (counter) part by using both `time_or_seq` and `seq_or_node`. In
    other words, it is important to ensure the monoticity of the UUIDv8 yourself.
    """

    def generate(
        self,
        timestamp_32: int,
        timestamp_48: int,  # Confusing. This is only 16 bit.
        time_or_seq: int,
        seq_or_node: int,
        node: int,
    ) -> UUID:
        """Generate UUID8.

        Args:
            timestamp_32: 32 bits timestamp.
            timestamp_48: 16 bits timestamp.
            time_or_seq: 12 bits timestamp or sequence.
            seq_or_node: 8 bits sequence or node.
            node: 54 bits node.

        Returns:
            UUIDv8.
        """
        # 32 bits
        timestamp_32 &= 0xFFFF_FFFF
        # 16 bits
        timestamp_48 &= 0xFFFF
        # 4 bits
        ver = 8
        # 12 bits
        time_or_seq &= 0x0FFF
        # 2 bits
        var = 0b10
        # 8 bits
        seq_or_node &= 0xFF
        # 54 bits
        node &= 0x3FFF_FFFF_FFFF

        uuid8_int = (
            (timestamp_32 << 96)
            | (timestamp_48 << 80)
            | (ver << 76)
            | (time_or_seq << 64)
            | (var << 62)
            | (seq_or_node << 54)
            | node
        )
        return UUID(int=uuid8_int)

    @classmethod
    @nodoc
    def parse(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        timestamp_32 = (uuid.int >> 96) & 0xFFFF_FFFF
        timestamp_48 = (uuid.int >> 80) & 0xFFFF
        version = (uuid.int >> 76) & 0xF
        assert version == 8, f"version must be 8, but {version}"
        time_or_seq = (uuid.int >> 64) & 0xFFF
        variant = (uuid.int >> 62) & 0x3
        seq_or_node = (uuid.int >> 54) & 0xFF
        node = uuid.int & 0x3FFF_FFFF_FFFF

        return {
            "timestamp_32": timestamp_32,
            "timestamp_48": timestamp_48,
            "ver": str(version),
            "time_or_seq": time_or_seq,
            "variant": variant,
            "seq_or_node": seq_or_node,
            "node": node,
        }


_uuid8_generator = UUID8Generator()


def uuid8(
    timestamp_32: int,
    timestamp_48: int,  # Don't get me wrong. This is only 16 bit.
    time_or_seq: int,
    seq_or_node: int,
    node: int,
) -> UUID:
    """Generate UUID8.

    Args:
        timestamp_32: 32 bits timestamp.
        timestamp_48: 16 bits timestamp.
        time_or_seq: 12 bits timestamp or sequence.
        seq_or_node: 8 bits sequence or node.
        node: 54 bits node.

    Returns:
        UUIDv8.
    """
    return _uuid8_generator.generate(timestamp_32, timestamp_48, time_or_seq, seq_or_node, node)



import time
from datetime import datetime, timedelta
from typing import Tuple


class UUIDClock:
    """Clock for UUID."""

    GREGORIAN_OFFSET = 0x1B2_1DD_213_814_000

    def time_ns(self) -> int:
        """Get time in nanoseconds.

        Returns:
            int: Epoch time in nanoseconds.
        """
        return time.time_ns()

    # ----------------------------
    # epoch_*_seconds
    # ----------------------------

    def epoch_seconds(self) -> int:
        """Get time in seconds.

        Returns:
            int: Epoch time in seconds.
        """
        return self.time_ns() // 1_000_000_000

    def epoch_milli_seconds(self) -> int:
        """Get time in milliseconds.

        Returns:
            int: Epoch time in milliseconds.
        """
        return self.time_ns() // 1_000_000

    def epoch_100_nano_seconds(self) -> int:
        """Get time in 100 nanoseconds.

        Returns:
            int: Epoch time in 100 nanoseconds.
        """
        return self.time_ns() // 100

    # ----------------------------
    # gregorian
    # ----------------------------

    def gregorian_100_nano_seconds(self) -> int:
        """Get time in 100 nanoseconds since Gregorian epoch.

        Returns:
            int: Time in 100 nanoseconds since Gregorian epoch.
        """
        return self.time_ns() // 100 + self.GREGORIAN_OFFSET

    # ----------------------------
    # epoch_36_bits_seconds_with_*_bits_*_seconds
    # ----------------------------

    def epoch_36_bits_seconds_with_12_bits_milli_seconds(self) -> int:
        """Get time in 12 bits milliseconds since epoch.

        Returns:
            int: Time in 12 bits milliseconds since epoch.
        """
        seconds, milli_part = divmod(self.time_ns() // 1_000_000, 1_000)
        return ((seconds & 0x000F_FFFF_FFFF) << 12) | milli_part

    def epoch_36_bits_seconds_with_24_bits_micro_seconds(self) -> int:
        """Get time in 24 bits microseconds since epoch.

        Returns:
            int: Time in 24 bits microseconds since epoch.
        """
        seconds, micro_part = divmod(self.time_ns() // 1_000, 1_000_000)
        return ((seconds & 0x000F_FFFF_FFFF) << 24) | micro_part

    def epoch_36_bits_seconds_with_38_bits_nano_seconds(self) -> int:
        """Get time in 38 bits nanoseconds since epoch.

        Returns:
            int: Time in 38 bits nanoseconds since epoch.
        """
        seconds, nano_part = divmod(self.time_ns(), 1_000_000_000)
        return ((seconds & 0x000F_FFFF_FFFF) << 38) | nano_part

    # ----------------------------
    # converter
    # ----------------------------

    @classmethod
    def to_datetime_from_epoch_seconds(cls, epoch_seconds: int) -> datetime:
        """Convert epoch seconds to datetime.

        Args:
            epoch_seconds (int): Epoch seconds.

        Returns:
            datetime: Datetime.
        """
        return datetime.fromtimestamp(epoch_seconds)

    @classmethod
    def to_datetime_from_epoch_milli_seconds(cls, epoch_milli_seconds: int) -> datetime:
        """Convert epoch milliseconds to datetime.

        Args:
            epoch_milli_seconds (int): Epoch milliseconds.

        Returns:
            datetime: Datetime.
        """
        return datetime.fromtimestamp(epoch_milli_seconds / 1_000)

    @classmethod
    def to_datetime_from_epoch_micro_seconds(cls, epoch_micro_seconds: int) -> datetime:
        """Convert epoch microseconds to datetime.

        Args:
            epoch_micro_seconds (int): Epoch microseconds.

        Returns:
            datetime: Datetime.
        """
        return datetime.fromtimestamp(epoch_micro_seconds / 1_000_000)

    @classmethod
    def to_datetime_from_epoch_100_nano_seconds(
        cls, epoch_100_nano_seconds: int
    ) -> Tuple[datetime, int]:
        """Convert epoch 100 nanoseconds to datetime.

        Args:
            epoch_100_nano_seconds (int): Epoch 100 nanoseconds.

        Returns:
            Tuple[datetime, int]: Datetime and nanoseconds.
        """
        return (
            datetime.fromtimestamp(epoch_100_nano_seconds / 10_000_000_000),
            epoch_100_nano_seconds % 10_000_000_000,
        )

    @classmethod
    def to_datetime_from_gregorian_100_nano_seconds(
        cls, gregorian_100_nano_seconds: int
    ) -> Tuple[datetime, int]:
        """Convert gregorian 100 nanoseconds to datetime.

        Args:
            gregorian_100_nano_seconds (int): Gregorian 100 nanoseconds.

        Returns:
            Tuple[datetime, int]: Datetime and nanoseconds.
        """
        epoch_100_nano_seconds = gregorian_100_nano_seconds - cls.GREGORIAN_OFFSET
        (seconds, fraction_nano) = divmod(epoch_100_nano_seconds * 100, 1_000_000_000)
        (microseconds, fraction_nano) = divmod(fraction_nano, 1000)
        return (
            datetime.fromtimestamp(seconds) + timedelta(microseconds=microseconds),
            fraction_nano,
        )

    @classmethod
    def to_datetime_from_epoch_nano_seconds(cls, epoch_nano_seconds: int) -> Tuple[datetime, int]:
        """Convert epoch nanoseconds to datetime.

        Args:
            epoch_nano_seconds (int): Epoch nanoseconds.

        Returns:
            Tuple[datetime, int]: Datetime and nanoseconds.
        """
        return (
            datetime.fromtimestamp(epoch_nano_seconds / 10_000_000_000),
            epoch_nano_seconds % 1_000_000_000,
        )



from typing import Any, Dict
from uuid import UUID

from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator

_nil_uuid = UUID("00000000-0000-0000-0000-000000000000")


class NilUUIDGenerator(UUIDGenerator):
    """Nil UUID generator."""

    def generate(self) -> UUID:
        """Generate Nil UUID."""
        return _nil_uuid

    @classmethod
    @nodoc
    def parse(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        return {
            "ver": "nil",
        }


_nil_uuid_generator = NilUUIDGenerator()


def nil_uuid() -> UUID:
    """Generate Nil UUID.

    Returns:
        UUID: Nil UUID.
    """
    return _nil_uuid_generator.generate()



import argparse
import json
import sys
from typing import List

from newnewid import (
    METHOD_0_NO_COUNTER,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_26,
    METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_42,
    METHOD_2_MONOTONIC_RANDOM_62_BITS,
    UUID,
    UUID7Option,
    UUIDGenerator,
    max_uuid,
    nil_uuid,
    uuid1,
    uuid3,
    uuid4,
    uuid5,
    uuid6,
    uuid7,
    uuid8,
)
from newnewid.parser.uuid_parser import parse


def _get_uuid7_option(method: str) -> UUID7Option:
    if method == "METHOD-0":
        return METHOD_0_NO_COUNTER
    elif method == "METHOD-1-12":
        return METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12
    elif method == "METHOD-1-26":
        return METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_26
    elif method == "METHOD-42":
        return METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_42
    elif method == "METHOD-2":
        return METHOD_2_MONOTONIC_RANDOM_62_BITS
    else:
        raise ValueError(f"Invalid method: {method}")


def generate_uuid(args):
    """Generate UUID."""
    uuid_version = args.uuid_version

    if uuid_version == "1":

        def uuid1_wrapper():
            return uuid1()

        uuid_module = uuid1_wrapper
    elif uuid_version == "3":
        namespace = args.namespace
        name = args.name
        if namespace is None or name is None:
            raise ValueError("namespace and name are required for UUIDv3")

        def uuid3_wrapper():
            return uuid3(UUID(namespace), name)

        uuid_module = uuid3_wrapper
    elif uuid_version == "4":
        uuid_module = uuid4
    elif uuid_version == "5":
        namespace = args.namespace
        name = args.name
        if namespace is None or name is None:
            raise ValueError("namespace and name are required for UUIDv5")

        def uuid5_wrapper():
            return uuid5(UUID(namespace), name)

        uuid_module = uuid5_wrapper
    elif uuid_version == "6":
        uuid_module = uuid6
    elif uuid_version == "7":
        uuid7_option = _get_uuid7_option(args.method)

        def uuid7_wrapper():
            return uuid7(uuid7_option)

        uuid_module = uuid7_wrapper
    elif uuid_version == "8":
        if args.custom_a is None or args.custom_b is None or args.custom_c is None:
            raise ValueError("custom-a, custom-b and custom-c are required for UUIDv8")

        def uuid8_wrapper():
            return uuid8(args.custom_a, args.custom_b, args.custom_c)

        uuid_module = uuid8_wrapper
    elif uuid_version == "nil":
        uuid_module = nil_uuid
    elif uuid_version == "max":
        uuid_module = max_uuid
    else:
        print("Invalid version.")
        sys.exit(1)

    for _ in range(args.n):
        print(uuid_module())


def parse_uuid(args):
    """Parse UUID."""
    uuid_strs: List[str] = args.uuid

    for uuid_str in uuid_strs:
        uuid = UUID(uuid_str)
        version = UUIDGenerator.get_version(uuid)
        if version == "7":
            uuid7_option = _get_uuid7_option(args.method)
            parsed = parse(uuid, uuid7_option=uuid7_option)
        else:
            parsed = parse(uuid)
        obj = {**{"uuid": uuid_str}, **parsed}
        print(json.dumps(obj))


def cli():
    """CLI."""
    parser = argparse.ArgumentParser(description="NewNewID")

    subparsers = parser.add_subparsers(required=True)
    gen_parser = subparsers.add_parser("generate", help="Generate UUID.")

    gen_parser.add_argument(
        "n",
        type=int,
        nargs="?",
        help="Number of UUIDs to generate. example: 10",
        default=1,
    )
    gen_parser.add_argument(
        "-u",
        "--uuid-version",
        type=str,
        help="Version of UUID. 1, 3, 4, 5, 6, 7, 8, nil or max",
        required=True,
    )
    gen_parser.add_argument(
        "--namespace",
        type=str,
        nargs="?",
        help="Namespace for UUIDv3 or UUIDv5. example: '0186b748-053d-7c24-be59-de03734bb9ab'",
        default=None,
    )
    gen_parser.add_argument(
        "--name",
        type=str,
        nargs="?",
        help="Name for UUIDv3 or UUIDv5. example: 'name'",
        default=None,
    )
    gen_parser.add_argument(
        "-m",
        "--method",
        nargs="?",
        type=str,
        help="Method for UUIDv7. 'METHOD-0', 'METHOD-1-12', 'METHOD-1-26', 'METHOD-1-42' or 'METHOD-2'. Default: 'METHOD-1-12'",
        default="METHOD-1-12",
    )
    gen_parser.add_argument(
        "--custom-a",
        type=int,
        nargs="?",
        help="cuatom-a binary for UUIDv8.",
        default=None,
    )
    gen_parser.add_argument(
        "--custom-b",
        type=int,
        nargs="?",
        help="cuatom-b binary for UUIDv8.",
        default=None,
    )
    gen_parser.add_argument(
        "--custom-c",
        type=int,
        nargs="?",
        help="cuatom-c binary for UUIDv8.",
        default=None,
    )

    def generate_uuid_wrapper(args):
        try:
            generate_uuid(args)
        except Exception:
            gen_parser.print_help()

    gen_parser.set_defaults(func=generate_uuid_wrapper)

    parse_parser = subparsers.add_parser("parse", help="Parse UUID.")
    parse_parser.add_argument("uuid", nargs="+", type=str, help="UUID")
    parse_parser.add_argument(
        "-m",
        "--method",
        nargs="?",
        type=str,
        help="Method for UUIDv7. 'METHOD-0', 'METHOD-1-12', 'METHOD-1-26', 'METHOD-1-42' or 'METHOD-2'. Default: 'METHOD-1-12'",
        default="METHOD-1-12",
    )

    def parse_uuid_wrapper(args):
        try:
            parse_uuid(args)
        except Exception:
            parse_parser.print_help()

    parse_parser.set_defaults(func=parse_uuid_wrapper)

    try:
        args = parser.parse_args()
        args.func(args)
    except Exception:
        parser.print_help()



from typing import Any, Dict
from uuid import UUID

from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator

_max_uuid = UUID("FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF")


class MaxUUIDGenerator(UUIDGenerator):
    """Max UUID generator."""

    def generate(self) -> UUID:
        """Generate Max UUID."""
        return _max_uuid

    @classmethod
    @nodoc
    def parse(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        return {
            "ver": "max",
        }


_max_uuid_generator = MaxUUIDGenerator()


def max_uuid() -> UUID:
    """Generate Max UUID.

    Returns:
        UUID: Max UUID.
    """
    return _max_uuid_generator.generate()



from newnewid.random.pseudo_random_generator import PseudoRandomGenerator


class PseudoRandomBinaryGenerator:
    """Pseudo random binary generator."""

    def __init__(
        self,
        mask_bits: int,
        pseudo_random_generator: PseudoRandomGenerator,
    ) -> None:
        """Create pseudo random binary generator.

        Args:
            mask_bits (int): Number of bits of the mask.
            pseudo_random_generator (PseudoRandomGenerator): Pseudo random generator.
        """
        assert 0 <= mask_bits, f"mask_bits must be greater than or equal to 0, not {mask_bits}"
        self.mask_bits = mask_bits
        self._pseudo_random_generator = pseudo_random_generator

    def generate(self) -> int:
        """Generate binary.

        Returns:
            int: Binary.
        """
        if self.mask_bits == 0:
            return 0

        return self._pseudo_random_generator.generate(self.mask_bits, "random-binary")



from abc import ABCMeta, abstractmethod
from threading import Lock
from time import sleep
from typing import Optional
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.exception.backward_uuid_exception import BackwardUUIDException
from newnewid.random.pseudo_random_generator import PseudoRandomGenerator
from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.uuid_generator import UUIDGenerator


class ClockBasedUUIDGenerator(
    UUIDGenerator,
    metaclass=ABCMeta,
):
    """UUID generator that uses clock."""

    def __init__(
        self,
        raise_exception_on_backward: bool,
        clock: Optional[UUIDClock],
        pseudo_random_generator: Optional[PseudoRandomGenerator],
        last_uuid: Optional[UUID],
    ) -> None:
        """Create ClockBasedUUIDGenerator.

        Args:
            raise_exception_on_backward (bool): Raise exception when backward UUID is generated if True. Otherwise, sleep until next UUID can be generated.
            clock (Optional[UUIDClock]): Clock.
            pseudo_random_generator (Optional[PseudoRandomGenerator]): Pseudo random generator.
            last_uuid (Optional[UUID]): Last generated UUID.
        """
        super().__init__()
        self.raise_exception_on_backward = raise_exception_on_backward
        self._clock = clock or UUIDClock()
        self._pseudo_random_generator = pseudo_random_generator or PseudoRandomGenerator()
        self.last_generated = last_uuid
        self._mask: Optional[int] = None
        self._lock = Lock()

    @abstractmethod
    def timestamp(self, clock: UUIDClock) -> int:
        """Get current timestamp.

        Args:
            clock (UUIDClock): Clock.

        Returns:
            int: Current timestamp.
        """

    @property
    @abstractmethod
    def clock_bits_length(self) -> int:
        """Get clock bits length.

        Returns:
            int: Clock bits length.
        """

    @property
    @abstractmethod
    def least_seconds(self) -> float:
        """Get least seconds in UUID.

        Returns:
            float: Least seconds.
        """

    @nodoc
    def generate(self) -> UUID:
        with self._lock:
            while True:
                if self._mask is None:
                    self._mask = (1 << self.clock_bits_length) - 1

                uuid = self.generate_impl(self.timestamp(self._clock) & self._mask)

                # backward UUID check
                if self.last_generated is not None and uuid <= self.last_generated:
                    if self.raise_exception_on_backward:
                        raise BackwardUUIDException(self.last_generated, uuid)
                    else:
                        sleep(self.least_seconds)
                        continue

                self.last_generated = uuid
                return uuid

    @abstractmethod
    def generate_impl(self, timestamp: int) -> UUID:
        """Generate UUID.

        Args:
            timestamp (int): Current timestamp.

        Returns:
            UUID: UUID.
        """



from typing import Any, Dict
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.gregorian_based_uuid_generator import (
    GregorianBasedUUIDGenerator,
)


class UUID6Generator(GregorianBasedUUIDGenerator):
    """UUIDv6 generator class.

    UUIDv6 is a 128-bit UUID that is based on Gregorian calendar time and is sortable.

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                           time_high                           |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |           time_mid            |  ver  |       time_low        |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |var|         clock_seq         |             node              |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                              node                             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    There are two variations of UUIDv6:

    * [Recommended] Pseudo-random node: pseudo-random number is used for `node`.
    * [NOT Recommended] MAC address node: MAC address is used for `node`.

    """

    @nodoc
    def build(
        self,
        time: int,
        ver: int,
        var: int,
        clock_seq: int,
        node: int,
    ) -> UUID:
        # 32 bits
        time_high = (time >> 28) & 0xFFFF_FFFF
        # 16 bits
        time_mid = (time >> 12) & 0xFFFF
        # lower 12 bits
        time_low = time & 0x0FFF

        return UUID(
            int=(time_high << 96)
            | (time_mid << 80)
            | (ver << 76)
            | (time_low << 64)
            | (var << 62)
            | (clock_seq << 48)
            | node
        )

    @classmethod
    @nodoc
    def parse(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        # 32 bits
        time_high = (uuid.int >> 96) & 0xFFFF_FFFF

        # 16 bits
        time_mid = (uuid.int >> 80) & 0xFFFF

        # 4 bits
        version = (uuid.int >> 76) & 0xF
        assert version == 6, f"version must be 6, but {version}"

        # 12 bits
        time_low = (uuid.int >> 64) & 0x0FFF

        # 2 bits
        variant = (uuid.int >> 62) & 0x3

        # 14 bits
        clock_seq = (uuid.int >> 48) & 0x3FFF

        # 48 bits
        node = uuid.int & 0xFFFF_FFFF_FFFF

        gregorian_100_nano_seconds = (time_high << 28) | (time_mid << 12) | time_low
        time, epoch_nano_fraction = UUIDClock.to_datetime_from_gregorian_100_nano_seconds(
            gregorian_100_nano_seconds
        )

        return {
            "time_high": time_high,
            "time_mid": time_mid,
            "ver": str(version),
            "time_low": time_low,
            "variant": variant,
            "clock_seq": clock_seq,
            "node": node,
            "gregorian_100_nano_seconds": gregorian_100_nano_seconds,
            "time": time.isoformat(),
            "epoch_nano_fraction": epoch_nano_fraction,
        }


_uses_mac_address_to_uuid6_generator: Dict[int, UUID6Generator] = {}


def uuid6(uses_mac_address: bool = False) -> UUID:
    """Generate UUIDv6.

    Args:
        uses_mac_address (bool, optional): MAC address is used for `node` if True. Otherwise, pseudo-random number is used. Defaults to False.
    Returns:
        UUID: UUIDv6.
    """
    if uses_mac_address not in _uses_mac_address_to_uuid6_generator:
        _uses_mac_address_to_uuid6_generator[uses_mac_address] = UUID6Generator(
            uses_mac_address=uses_mac_address
        )

    return _uses_mac_address_to_uuid6_generator[uses_mac_address].generate()



from abc import ABCMeta, abstractmethod
from typing import Any, Dict
from uuid import UUID


class UUIDGenerator(metaclass=ABCMeta):
    """UUID generator."""

    @abstractmethod
    def generate(self) -> UUID:
        """Generate UUID.

        Returns:
            UUID: UUID.
        """

    @classmethod
    @abstractmethod
    def parse(cls, uuid: UUID, **args: Any) -> Dict[str, Any]:
        """Parse UUID.

        Args:
            uuid (UUID): UUID

        Returns:
            Dict[str, Any]: Parsed UUID.
        """

    @classmethod
    def get_version(cls, uuid: UUID) -> str:
        """Get version of UUID.

        Args:
            uuid (UUID): UUID

        Returns:
            str: Version of UUID.
        """
        uuid_int = uuid.int
        version = int((uuid_int >> 76) & 0xF)
        if version == 0:
            return "nil"
        elif version == 0b1111:
            return "max"
        else:
            return str(version)



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, Optional, Tuple
from uuid import UUID

from newnewid.clock.uuid_clock import UUIDClock
from newnewid.counter.counter import Counter
from newnewid.random.pseudo_random_binary_generator import PseudoRandomBinaryGenerator
from newnewid.random.pseudo_random_generator import PseudoRandomGenerator
from newnewid.util.nodoc import nodoc
from newnewid.uuidgenerator.clock_based_uuid_generator import ClockBasedUUIDGenerator


@dataclass(frozen=True)
class UUID7Option:
    """Option for UUIDv7."""

    counter_bits_length: int
    increment_bits_length: Optional[int]

    @nodoc
    def __post_init__(self):
        assert (
            0 <= self.counter_bits_length <= 74
        ), f"counter_bits_length must be between 0 and 74, not {self.counter_bits_length}"

        if self.increment_bits_length is not None:
            assert (
                self.increment_bits_length <= self.counter_bits_length
            ), f"increment_bits_length must be equal or less than counter_bits_length, not {self.increment_bits_length}"

    @classmethod
    def method_1_fixed_length_dedicated_counter_bits(
        cls, counter_bits_length: int
    ) -> "UUID7Option":
        """Create option for UUIDv7 of method 1.

        Args:
            counter_bits_length (int): Counter value length in bits.

        Returns:
            UUID7Option: Option for UUIDv7 of method 1.
        """
        assert (
            12 <= counter_bits_length <= 42
        ), f"counter_bits_length must be between 12 and 42, not {counter_bits_length}"

        return cls(
            counter_bits_length=counter_bits_length,
            increment_bits_length=1,
        )

    @classmethod
    def method_2_monotonic_random(cls, max_increment_bits_length: int) -> "UUID7Option":
        """Create option for UUIDv7 of method 2.

        Args:
            max_increment_bits_length (int): Maximum increment value length in bits.
            Range of the value is 1 to 2^max_increment_bits_length - 1.

        Returns:
            UUID7Option: Option for UUIDv7 of method 2.
        """
        assert (
            1 <= max_increment_bits_length <= 74
        ), f"increment_bits_length must be between 1 and 72, not {max_increment_bits_length}"

        return cls(
            counter_bits_length=74,
            increment_bits_length=max_increment_bits_length,
        )


METHOD_0_NO_COUNTER = UUID7Option(counter_bits_length=0, increment_bits_length=None)

METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12 = (
    UUID7Option.method_1_fixed_length_dedicated_counter_bits(
        counter_bits_length=12,
    )
)

METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_26 = (
    UUID7Option.method_1_fixed_length_dedicated_counter_bits(
        counter_bits_length=26,
    )
)

METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_42 = (
    UUID7Option.method_1_fixed_length_dedicated_counter_bits(
        counter_bits_length=42,
    )
)


METHOD_2_MONOTONIC_RANDOM_62_BITS = UUID7Option.method_2_monotonic_random(
    max_increment_bits_length=62,
)


_ULID_COMPATIBLE_OPTION = UUID7Option.method_2_monotonic_random(
    max_increment_bits_length=1,
)


class UUID7Generator(ClockBasedUUIDGenerator):
    """UUIDv7 generator class.

    UUIDv7 is a 128-bit UUID that is based on Epoch time and is sortable.

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                           unix_ts_ms                          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |          unix_ts_ms           |  ver  |       rand_a          |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |var|                        rand_b                             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                            rand_b                             |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    There are two variations of UUIDv7:

    * Method 1: Fixed-length dedicated counter bits: `rand_a` and `rand_b` are used as counters,
      with a length of 1 to 74 bits. When `unix_ts_ms` is the same as the previous timestamp,
      the counter is incremented by 1. The counter is reset to a random number when `unix_ts_ms`
      changes. The remaining 73 to 1 bits of `rand_a` and `rand_b` are used for pseudo-random
      numbers.
    * Method 2: Monotonic random: `rand_a` and `rand_b` are used to generate random numbers.
      This method is a special case of Method 1, with a counter length of 74 bits, incremented
      by a pseudo-random number. The number is generated from a range of values that can be
      represented by a fixed-length bit sequence, excluding 0.

    The parameter for Method 1 is `counter_bits_length`, which is the length of the counter bits.
    The parameter for Method 2 is `increment_bits_length`, which is the bit length that determines
    the range of pseudo-random values generated during incrementation.

    Method 1 with counter_bits_length=74 is equivalent to Method 2 with increment_bits_length=1.
    """

    def __init__(
        self,
        uuid7_option: UUID7Option,
        raise_exception_on_backward: bool = False,
        clock: Optional[UUIDClock] = None,
        pseudo_random_generator: Optional[PseudoRandomGenerator] = None,
        last_uuid: Optional[UUID] = None,
    ) -> None:
        """Create UUIDv7 generator.

        Args:
            uuid7_option (UUID7Option): UUIDv7 option.
            raise_exception_on_backward (bool, optional): Raise exception when backward UUID is generated. Otherwise, sleep until the next milliseconds. Defaults to False.
            clock (Optional[UUIDClock], optional): Clock. Defaults to None.
            pseudo_random_generator (Optional[PseudoRandomGenerator], optional): Pseudo random generator. Defaults to None.
            last_uuid (Optional[UUID], optional): Last UUID. Defaults to None.
        """
        super().__init__(
            raise_exception_on_backward=raise_exception_on_backward,
            clock=clock,
            pseudo_random_generator=pseudo_random_generator,
            last_uuid=last_uuid,
        )

        last_timestamp: Optional[int] = None
        last_counter: Optional[int] = None
        if last_uuid:
            assert last_uuid.version == 7, f"last_uuid must be version 7, not {last_uuid.version}"

            last_timestamp, last_counter = self.__class__._parse_last_uuid(
                last_uuid,
                uuid7_option,
            )

        self._counter = Counter(
            bits_length=uuid7_option.counter_bits_length,
            max_increment_bits_length=uuid7_option.increment_bits_length,
            pseudo_random_generator=self._pseudo_random_generator,
            initial_timestamp=last_timestamp,
            initial_counter=last_counter,
        )

        assert (
            0 <= self._counter.bits_length <= 74
        ), f"mask_bits_length must be between 0 and 74, not {self._counter.bits_length}"
        self.rand_bits = 74 - self._counter.bits_length

        self._pseudo_random_binary_generator = PseudoRandomBinaryGenerator(
            self.rand_bits, self._pseudo_random_generator
        )

    @nodoc
    def timestamp(self, clock: UUIDClock) -> int:
        return clock.epoch_milli_seconds()

    @property
    @nodoc
    def clock_bits_length(self) -> int:
        return 48

    @property
    @nodoc
    def least_seconds(self) -> float:
        return 0.001

    @nodoc
    def generate_impl(self, timestamp: int) -> UUID:
        # 48 bits
        unix_ts_ms = timestamp

        # 4 bits
        ver = 7

        # 0 bits - 74 bits
        seq = self._counter.get_next(unix_ts_ms)
        # 74 bits - 0 bits
        node = self._pseudo_random_binary_generator.generate()

        # 74 bits
        seq_and_node = (seq << self.rand_bits) | node

        # 12 bits = 74 - 62
        rand_a = (seq_and_node >> 62) & 0x0FFF

        # 62 bits
        rand_b = seq_and_node & 0x3FFF_FFFF_FFFF_FFFF

        # 2 bits
        var = 0b10

        uuid_int = (unix_ts_ms << 80) | (ver << 76) | (rand_a << 64) | (var << 62) | rand_b
        return UUID(int=uuid_int)

    @classmethod
    def parse(cls, uuid: UUID, uuid7_option: UUID7Option, **kwargs: Any) -> Dict[str, Any]:
        """Parse UUIDv7.

        Args:
            uuid (UUID): UUID.
            uuid7_option (UUID7Option): UUIDv7 option.

        Returns:
            Dict[str, Any]: Parsed UUIDv7.
        """
        # 48 bits
        unix_ts_ms = (uuid.int >> 80) & 0xFFFF_FFFF_FFFF

        # 4 bits
        ver = (uuid.int >> 76) & 0x0F
        assert ver == 7, f"version must be 7, not {ver}"

        # 12 bits
        rand_a = (uuid.int >> 64) & 0x0FFF

        # 2 bits
        variant = (uuid.int >> 62) & 0x03

        # 62 bits
        rand_b = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

        # 0 - 74 bits
        seq, rand = cls._parse_counter_and_rand(rand_a, rand_b, uuid7_option)

        time = datetime.fromtimestamp(unix_ts_ms / 1000)
        return {
            "unix_ts_ms": unix_ts_ms,
            "ver": "7",
            "rand_a": rand_a,
            "var": variant,
            "rand_b": rand_b,
            "time": time.isoformat(),
            "seq": seq,
            "rand": rand,
        }

    @classmethod
    def _parse_counter_and_rand(
        cls,
        rand_a: int,
        rand_b: int,
        uuid7_option: UUID7Option,
    ) -> Tuple[Optional[int], Optional[int]]:
        seq_and_rand = (rand_a << 62) | rand_b
        if uuid7_option.counter_bits_length == 0:
            return None, seq_and_rand
        elif uuid7_option.counter_bits_length == 74:
            return seq_and_rand, None

        seq_mask = (1 << uuid7_option.counter_bits_length) - 1
        rand_bits = 74 - uuid7_option.counter_bits_length
        rand_mask = (1 << rand_bits) - 1

        seq = (seq_and_rand >> rand_bits) & seq_mask
        rand = seq_and_rand & rand_mask

        return seq, rand

    @classmethod
    def _parse_last_uuid(
        cls,
        last_uuid: UUID,
        uuid7_option: UUID7Option,
    ) -> Tuple[int, Optional[int]]:
        parsed = cls.parse(last_uuid, uuid7_option)
        unix_ts_ms: int = parsed["unix_ts_ms"]
        counter, _ = cls._parse_counter_and_rand(
            rand_a=parsed["rand_a"],
            rand_b=parsed["rand_b"],
            uuid7_option=uuid7_option,
        )

        return (unix_ts_ms, counter)


_option_to_uuid7_generator: Dict[UUID7Option, UUID7Generator] = {}


def uuid7(
    uuid7_option: Optional[UUID7Option] = None,
) -> UUID:
    """Generate UUIDv7.

    Args:
        uuid7_option (Optional[UUID7Option], optional): UUIDv7 option. Defaults to None.

    Returns:
        UUID: UUIDv7.
    """
    uuid7_option = uuid7_option or METHOD_1_FIXED_LENGTH_DEDICATED_COUNTER_BITS_12
    if uuid7_option not in _option_to_uuid7_generator:
        _option_to_uuid7_generator[uuid7_option] = UUID7Generator(
            uuid7_option=uuid7_option,
        )
    uuid7_generator = _option_to_uuid7_generator[uuid7_option]

    return uuid7_generator.generate()


# ==================== BONUS TRACK ====================

_ulid_compatible_generator = UUID7Generator(_ULID_COMPATIBLE_OPTION)


def ulid_compatible() -> UUID:
    """Generate ULID compatible UUIDv7.

    Returns:
        UUID: UUIDv7.
    """
    return _ulid_compatible_generator.generate()



from typing import Optional

from newnewid.random.pseudo_random_generator import PseudoRandomGenerator


class Counter:
    """Timebased counter.

    This class manages a counter.

    The counter is incremented when the same timestamp is given and reset when the timestamp is changed.
    """

    def __init__(
        self,
        bits_length: int,
        max_increment_bits_length: Optional[int],
        pseudo_random_generator: PseudoRandomGenerator,
        initial_timestamp: Optional[int],
        initial_counter: Optional[int],
    ) -> None:
        """Create counter.

        Args:
            bits_length (int): Number of bits of the counter.
            max_increment_bits_length (Optional[int]): Maximum number of bits of the increment. Maximum value is 2^max_increment_bits_length - 1.
            pseudo_random_generator (PseudoRandomGenerator): Pseudo random generator.
            initial_last_timestamp (Optional[int]): Initial timestamp.
            initial_counter (Optional[int]): Initial counter value.
        """
        if bits_length > 0:
            assert max_increment_bits_length is not None, "increment_bits_length must be specified"
            assert (
                bits_length >= max_increment_bits_length
            ), f"mask_bits_length must be greater than or equal to increment_bits_length, not {bits_length} < {max_increment_bits_length}"
            assert (
                max_increment_bits_length >= 1
            ), f"increment_bits_length must be greater than or equal to 1, not {max_increment_bits_length}"

        self.bits_length = bits_length
        self.mask: int = (1 << bits_length) - 1
        self.max_increment_bits_length = max_increment_bits_length
        self._pseudo_random_generator = pseudo_random_generator

        self.last_timestamp = initial_timestamp
        if initial_counter is None:
            if bits_length > 0:
                self.counter = self._get_counter_reset_value()
            else:
                self.counter = 0
        else:
            self.counter = initial_counter

    def _get_counter_reset_value(self) -> int:
        return self._pseudo_random_generator.generate(self.bits_length - 1, "counter-reset")

    def get_next(self, timestamp: int) -> int:
        """Ge next counter value.

        Args:
            timestamp (int): Current timestamp.

        Returns:
            int: counter value.
        """
        if self.bits_length == 0:
            return 0

        if self.last_timestamp is not None and timestamp <= self.last_timestamp:
            if self.max_increment_bits_length == 1:
                increment = 1
            else:
                assert self.max_increment_bits_length is not None
                increment = 0
                while increment <= 0:
                    increment = self._pseudo_random_generator.generate(
                        self.max_increment_bits_length, "counter-increment"
                    )

            new_counter_raw = self.counter + increment
        else:
            new_counter_raw = self._get_counter_reset_value()
        self.counter = new_counter_raw & self.mask
        self.last_timestamp = timestamp
        return self.counter



from typing import Any, Dict
from uuid import UUID

from newnewid.util.nodoc import nodoc


class UUID8Generator:
    """UUIDv8 generator class.

     0                   1                   2                   3
     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                           custom_a                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |          custom_a             |  ver  |       custom_b        |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |var|                       custom_c                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    |                           custom_c                            |
    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
    """

    def generate(
        self,
        custom_a: int,
        custom_b: int,
        custom_c: int,
    ) -> UUID:
        """Generate UUIDv8.

        UUIDv8 can define 122 bits of information as desired.
        The 122 bits are stored in 48-bits, 12-bits, and 62-bits parts.

        Args:
            custom_a (int): 48-bits part
            custom_b (int): 12-bits part
            custom_c (int): 62-bits part

        Returns:
            UUID: UUIDv8
        """
        # 48 bits
        custom_a &= 0xFFFF_FFFF_FFFF
        # 4 bits
        ver = 8
        # 12 bits
        custom_b &= 0x0FFF
        # 2 bits
        var = 2
        # 62 bits
        custom_c &= 0x3FFF_FFFF_FFFF_FFFF

        return UUID(int=(custom_a << 80) | (ver << 76) | (custom_b << 64) | (var << 62) | custom_c)

    @classmethod
    @nodoc
    def parse(cls, uuid: UUID, **kwargs: Any) -> Dict[str, Any]:
        custom_a = (uuid.int >> 80) & 0xFFFF_FFFF_FFFF
        version = (uuid.int >> 76) & 0xF
        custom_b = (uuid.int >> 64) & 0x0FFF
        variant = (uuid.int >> 62) & 0x03
        custom_c = uuid.int & 0x3FFF_FFFF_FFFF_FFFF

        return {
            "ver": str(version),
            "custom_a": custom_a,
            "custom_b": custom_b,
            "variant": variant,
            "custom_c": custom_c,
        }


_uuid8_generator = UUID8Generator()


def uuid8(
    custom_a: int,
    custom_b: int,
    custom_c: int,
) -> UUID:
    """Generate UUIDv8.

    UUIDv8 can define 122 bits of information as desired.
    The 122 bits are stored in 48-bits, 12-bits, and 62-bits parts.

    Args:
        custom_a (int): 48-bits part
        custom_b (int): 12-bits part
        custom_c (int): 62-bits part

    Returns:
        UUID: UUIDv8
    """
    return _uuid8_generator.generate(custom_a, custom_b, custom_c)



# noqa: E401
import importlib
from typing import Literal

UUIDSpec = Literal[
    "draft_peabody_dispatch_new_uuid_format_01",
    "draft_peabody_dispatch_new_uuid_format_02",
    "draft_peabody_dispatch_new_uuid_format_03",
    "draft_peabody_dispatch_new_uuid_format_04",
    "draft_ietf_uuidrev_rfc4122bis_00",
    "draft_ietf_uuidrev_rfc4122bis_01",
    "draft_ietf_uuidrev_rfc4122bis_02",
    "latest",
]


def load_modules(spec: UUIDSpec = "latest"):
    """Load modules for the given spec.

    Args:
        spec (UUIDSpec, optional): Specification to load. Defaults to "latest".
    """
    if spec == "draft_peabody_dispatch_new_uuid_format_01":
        module = importlib.import_module("newnewid.draft_peabody_dispatch_new_uuid_format_01")
    elif spec == "draft_peabody_dispatch_new_uuid_format_02":
        module = importlib.import_module("newnewid.draft_peabody_dispatch_new_uuid_format_02")
    elif spec == "draft_peabody_dispatch_new_uuid_format_03":
        module = importlib.import_module("newnewid.draft_peabody_dispatch_new_uuid_format_03")
    elif spec == "draft_peabody_dispatch_new_uuid_format_04":
        module = importlib.import_module("newnewid.draft_peabody_dispatch_new_uuid_format_04")
    elif spec == "draft_ietf_uuidrev_rfc4122bis_00":
        module = importlib.import_module("newnewid.draft_ietf_uuidrev_rfc4122bis_00")
    elif spec == "draft_ietf_uuidrev_rfc4122bis_01":
        module = importlib.import_module("newnewid.draft_ietf_uuidrev_rfc4122bis_01")
    elif spec == "draft_ietf_uuidrev_rfc4122bis_02" or spec == "latest":
        module = importlib.import_module("newnewid.draft_ietf_uuidrev_rfc4122bis_02")
    else:
        raise ValueError(f"Unknown spec: {spec}")

    return module
